package ti.util;

import java.io.*;

/** Converts a TI disk track image to a sector image. This is required to 
    convert PC99 files to v9t9 files used in MESS. PC99 files contain the
    complete track information, so we need to extract the sector contents
    and glue them together in one stream. 
    In the track, sectors are numbered from 0 to 17 for 360K disk images. Tracks
    are numbered 0 to 39 on each side. Heads are numbered 0 or 1.
    
    Important note: Sectors are read starting in track 0 on side 0, going
    up to track 39 on side 0, but then continuing on track 39 on side 1, 
    going back down to track 0 on side 1.

    This tool currently only works with DSDD disk images.
    
    Michael Zapf, 2007-12-26    
*/
public class SectorDump {
    
    byte[] m_abyContent;
    
    public static final int TRACKHEAD  = 0x0010;
    public static final int SECTHEAD   = 0x0052;
    public static final int TRACKTAIL  = 0x02e0;
    public static final int SECTTAIL   = 0x0002;
    public static final int SECTLENGTH = 0x0100;
    public static final int SECTTOTAL = SECTLENGTH+SECTHEAD+SECTTAIL;
    
    public static final int TRACKSIZE = 18;
    public static final int TRACKS = 40;
    
    // 6872 bytes all per track
    // 4608 bytes sector data
    // 2264 bytes header per track
    // 296 bytes all per sector
    // 720 bytes sector header
    
    public static void main(String[] arg) {
        byte[] abyFile = null;
        try {
            FileInputStream fis = new FileInputStream(arg[0]);
            DataInputStream dis = new DataInputStream(fis);
            abyFile = new byte[dis.available()];
            dis.readFully(abyFile);
            
            FileOutputStream fos = new FileOutputStream(arg[1]);
            SectorDump sd = new SectorDump(abyFile);
            sd.convert(fos);
        }
        catch (FileNotFoundException fnfx) {
            System.err.println("File not found: " + fnfx.getMessage());
        }
        catch (IOException iox) {
            System.err.println("File unreadable: " + iox.getMessage());
        }
    }
    
    public SectorDump(byte[] abyContent) {
        m_abyContent = abyContent;
    }
    
    public void convert(FileOutputStream fos) throws IOException {
        int nPos = TRACKHEAD;
        int nMax = 0;
        int nSect = TRACKSIZE;
        
        byte[] abyTarget = new byte[1440*256];
        
        while (nPos < m_abyContent.length && nMax<1440) {
            int nHead = (m_abyContent[39 + nPos]&0xff);
            int nTrack =  (m_abyContent[38 + nPos]&0xff);
            int nSector = (m_abyContent[40 + nPos]&0xff); 

            int nSectorNumber = getSectorNumber(nHead, nTrack, nSector);
            System.arraycopy(m_abyContent, nPos + SECTHEAD, abyTarget, nSectorNumber*SECTLENGTH, SECTLENGTH);

            nPos += SECTTOTAL;
            nSect--;
            if (nSect==0) {
                nPos += TRACKTAIL + TRACKHEAD;
                nSect = TRACKSIZE;
            }
            nMax++;
        }
        
        fos.write(abyTarget);
        fos.close();
    }

    private int getSectorNumber(int nHead, int nTrack, int nSector) {
        // Note that tracks are read 0a, 1a, ... 38a, 39a, 39b, 38b, 37b, ... 0b 
        int nPosition = 0;
        if (nHead==0) {
            nPosition = nTrack * TRACKSIZE + nSector;
        }
        else {
            nPosition = TRACKS * TRACKSIZE + (TRACKS-1 - nTrack) * TRACKSIZE + nSector;
        }
        return nPosition;
    }
}
